import json
import uuid
from datetime import datetime
from typing import Dict, List, Optional

import pandas as pd
from answers.backend.db.messages import messages_sql_manager
from answers.backend.db.sql.dataset_schemas import ANSWERS_DATASETS_METADATA, CONVERSATION_DATASET_CONF_ID
from common.backend.db.sql.queries import (
    DeleteQueryBuilder,
    InsertQueryBuilder,
    UpdateQueryBuilder,
    WhereCondition,
    get_post_queries,
)
from common.backend.db.sql.tables_managers import GenericConversationSQL
from common.backend.models.base import ConversationInfo, ConversationInsertInfo, RecordState
from common.backend.utils.dataiku_api import dataiku_api
from common.backend.utils.sql_timing import log_query_time
from common.llm_assist.logging import logger
from dataiku.sql.expression import Operator
from werkzeug.exceptions import BadRequest


#TODO refactor this method as it is a duplicate of the same method in messages.py
def stringify(data) -> Optional[str]:
    return json.dumps(data, ensure_ascii=False) if data else None


class ConversationsSQL(GenericConversationSQL):
    def __init__(self):
        config = dataiku_api.webapp_config
        if not config.get("enable_answers_api", False) or not config.get(CONVERSATION_DATASET_CONF_ID):
            return
        super().__init__(config=config,
                         columns=ANSWERS_DATASETS_METADATA[CONVERSATION_DATASET_CONF_ID]["columns"],
                         dataset_conf_id=CONVERSATION_DATASET_CONF_ID,
                         default_project_key=dataiku_api.default_project_key)

        self.permanent_delete = self.config.get("permanent_delete", True)

    @log_query_time
    def add_conversation(
        self,
        user: str,
        conversation_info: ConversationInsertInfo,
    ):
        conversation_id = str(uuid.uuid4())  # generate
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        conversation_name = conversation_info["name"]
        record_value = [
            conversation_id,
            user,
            conversation_info.get("platform"),
            timestamp,
            conversation_name,
            timestamp,
            RecordState.PRESENT.value,
            stringify(conversation_info.get("metadata")),
        ]
        insert_query = (
            InsertQueryBuilder(self.dataset).add_columns(self.columns).add_values(values=[record_value]).build()
        )
        try:
            self.executor.query_to_df(insert_query, post_queries=get_post_queries(self.dataset))
        except Exception as err:
            logger.exception(err)
            raise BadRequest(f"Error when executing SQL query: {err}")
        return ConversationInfo(
            id=conversation_id, name=conversation_name, timestamp=timestamp, created_at=timestamp, updated_at=timestamp #type: ignore
        )

    @log_query_time
    def delete_conversation_permanently(self, conditions):
        delete_query = DeleteQueryBuilder(self.dataset).add_conds(conds=conditions).build()
        try:
            self.executor.query_to_df(delete_query, post_queries=get_post_queries(self.dataset))
        except Exception as err:
            logger.error(err)
            raise BadRequest(f"Error when executing SQL query: {err}")

    @log_query_time
    def update_conversation_as_deleted(self, conditions):
        update_query = (
            UpdateQueryBuilder(self.dataset)
            .add_set_cols([(self.col("state"), RecordState.CLEARED.value)])
            .add_conds(conditions)
            .build()
        )

        try:
            self.executor.query_to_df(update_query, post_queries=get_post_queries(self.dataset))
        except Exception as err:
            logger.error(err)
            raise BadRequest(f"Error when executing SQL query: {err}")

    @log_query_time
    def delete_conversation(self, user: str, conversation_id: str):
        conditions = [
            WhereCondition(
                column=self.col("conversation_id"),
                value=conversation_id,
                operator=Operator.EQ,
            ),
            WhereCondition(column=self.col("user"), value=user, operator=Operator.EQ),
        ]
        if self.permanent_delete:
            # first delete all the messages associated with the conversation
            messages_sql_manager.delete_conversation_messages(user=user, conversation_id=conversation_id)
            # then delete the conversation
            self.delete_conversation_permanently(conditions)
        else:
            self.update_conversation_as_deleted(conditions)

    @log_query_time
    def get_all_user_conversations(self, user: str, present_only: bool = True) -> List[ConversationInfo]:
        conditions = [
            WhereCondition(column=self.col("user"), value=user, operator=Operator.EQ),
        ]
        present_condition = None
        if present_only:
            present_condition = WhereCondition(
                column=self.col("state"),
                operator=Operator.OR,
                value=[RecordState.PRESENT.value, RecordState.DELETED.value],
            )
        else:
            present_condition = WhereCondition(
                column=self.col("state"), value=RecordState.PRESENT.value, operator=Operator.EQ
            )
        conditions.append(present_condition)
        format_ = "dataframe"
        order_by = self.col("updated_at")
        result: pd.DataFrame = self.select_columns_from_dataset(
            column_names=self.columns,
            distinct=True,
            eq_cond=conditions,
            format_=format_,
            order_by=order_by,
        )
        conversations: Dict[str, ConversationInfo] = {}
        for index, row in result.iterrows():
            conversation_id = row["conversation_id"]
            conversations[conversation_id] = ConversationInfo(
                id=conversation_id,
                name=row["conversation_name"],
                timestamp=row["created_at"],
                created_at=row["created_at"],
                updated_at=row["updated_at"],
            )
        return list(conversations.values())

conversations_sql_manager = ConversationsSQL()
