import itertools
import json
import uuid
from datetime import datetime
from typing import Generator, List, Optional

import pandas as pd
from answers.backend.db.sql.dataset_schemas import ANSWERS_DATASETS_METADATA, MESSAGE_DATASET_CONF_ID
from common.backend.db.sql.queries import (
    DeleteQueryBuilder,
    InsertQueryBuilder,
    UpdateQueryBuilder,
    WhereCondition,
    get_post_queries,
)
from common.backend.db.sql.tables_managers import GenericMessageSQL
from common.backend.models.base import MessageInsertInfo, RecordState
from common.backend.models.data import MessageInfo
from common.backend.utils.dataiku_api import dataiku_api
from common.backend.utils.file_utils import delete_files
from common.backend.utils.sql_timing import log_query_time
from common.llm_assist.logging import logger
from dataiku.sql.expression import Operator
from werkzeug.exceptions import BadRequest


def stringify(data) -> Optional[str]:
    return json.dumps(data, ensure_ascii=False) if data else None


class MessageSQL(GenericMessageSQL):
    def __init__(self):
        config = dataiku_api.webapp_config
        if not config.get("enable_answers_api", False):
            return
        super().__init__(config=config,
                         columns=ANSWERS_DATASETS_METADATA[MESSAGE_DATASET_CONF_ID]["columns"],
                         dataset_conf_id=MESSAGE_DATASET_CONF_ID,
                         default_project_key=dataiku_api.default_project_key
        )

        self.permanent_delete = self.config.get("permanent_delete", True)

    @log_query_time
    def add_message(
        self,
        user: str,
        message_info: MessageInsertInfo,
    ) -> str:
        message_id = str(uuid.uuid4())  # generate
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        record_value = [
            message_id,
            user,
            message_info.get("platform"),
            message_info.get("llm_name"),
            message_info.get("query"),
            message_info.get("answer"),
            stringify(message_info.get("filters")),
            stringify(message_info.get("sources")),
            timestamp,
            stringify(message_info.get("llm_context")),
            stringify(message_info.get("generated_media")),
            stringify(message_info.get("feedback")),
            message_info.get("conversation_id"),
            RecordState.PRESENT.value,
            stringify(message_info.get("history")),
        ]
        insert_query = (
            InsertQueryBuilder(self.dataset).add_columns(self.columns).add_values(values=[record_value]).build()
        )
        try:
            self.executor.query_to_df(insert_query, post_queries=get_post_queries(self.dataset))
        except Exception as err:
            logger.exception(err)
            raise BadRequest(f"Error when executing SQL query: {err}")
        return message_id

    @log_query_time
    def delete_messages_permanently(self, conditions: List[WhereCondition]):
        delete_query = DeleteQueryBuilder(self.dataset).add_conds(conds=conditions).build()
        try:
            self.executor.query_to_df(delete_query, post_queries=get_post_queries(self.dataset))
        except Exception as err:
            logger.error(err)
            raise BadRequest(f"Error when executing SQL query: {err}")

    @log_query_time
    def update_messages_as_deleted(self, conditions: List[WhereCondition]):
        update_query = (
            UpdateQueryBuilder(self.dataset)
            .add_set_cols([(self.col("state"), RecordState.DELETED.value)])
            .add_conds(conditions)
            .build()
        )

        try:
            self.executor.query_to_df(update_query, post_queries=get_post_queries(self.dataset))
        except Exception as err:
            logger.error(err)
            raise BadRequest(f"Error when executing SQL query: {err}")

    # for now copied from conversation.py - could be moved to another file
    @staticmethod
    def get_paths_from_history(series: pd.Series, data_key: str, additional_file_name: str) -> List[str]:
        def get_paths(row):
            if not row:
                return []
            media_qa_context = json.loads(row).get(data_key, [])
            return itertools.chain(
                (media.get(additional_file_name) for media in media_qa_context if media.get(additional_file_name)),
                (media.get("file_path") for media in media_qa_context if media.get("file_path")),
            )

        series = series.map(get_paths)  # type: ignore
        unique_values = list(set(item for sublist in series for item in sublist if item))
        return unique_values

    @log_query_time
    def get_message_media_paths(self, conditions: List[WhereCondition]) -> List[str]:
        if dataiku_api.webapp_config.get("upload_folder") is None:
            return []
        media_columns = [
            self.col("llm_context"),
            self.col("generated_media"),
        ]
        media_df = self.select_columns_from_dataset(
            column_names=media_columns,
            eq_cond=conditions,
        )
        if isinstance(media_df, Generator) or media_df.empty:
            return []

        llm_context = media_df["llm_context"].dropna()
        generated_media = media_df["generated_media"].dropna()
        uploaded = MessageSQL.get_paths_from_history(
            series=llm_context, data_key="media_qa_context", additional_file_name="metadata_path"
        )
        generated = MessageSQL.get_paths_from_history(
            series=generated_media, data_key="images", additional_file_name="referred_file_path"
        )
        return uploaded + generated

    def delete_files(self, conditions: List[WhereCondition]):
        # TODO handle permenant delete and not permenant delete
        media_to_delete = self.get_message_media_paths(conditions)
        if media_to_delete:
            logger.debug(f"Deleting media files: {media_to_delete}")
            delete_files(media_to_delete)

    @log_query_time
    def delete_message(self, user: str, message_id: str):
        conditions = [
            WhereCondition(
                column=self.col("message_id"),
                value=message_id,
                operator=Operator.EQ,
            ),
            WhereCondition(column=self.col("user"), value=user, operator=Operator.EQ),
            # TODO: does it make sense to add present_condition to all queries?
            WhereCondition(
                column=self.col("state"),
                operator=Operator.EQ,
                value=RecordState.PRESENT.value,
            )
        ]
        self.delete_files(conditions)
        if self.permanent_delete:
            self.delete_messages_permanently(conditions)
        else:
            self.update_messages_as_deleted(conditions)

    @log_query_time
    def get_all_conversation_messages(self, user: str, conversation_id: str, only_present: bool = True):
        conditions = [
            WhereCondition(
                column=self.col("conversation_id"),
                value=conversation_id,
                operator=Operator.EQ,
            ),
            WhereCondition(column=self.col("user"), value=user, operator=Operator.EQ),
        ]
        if only_present:
            conditions.append(
                WhereCondition(
                    column=self.col("state"),
                    value=RecordState.PRESENT.value,
                    operator=Operator.EQ,
                )
            )
        format_ = "dataframe"
        order_by = self.col("created_at")
        result: pd.DataFrame = self.select_columns_from_dataset(
            column_names=self.columns, eq_cond=conditions, format_=format_, order_by=order_by
        )
        messages: List[MessageInfo] = []
        for index, row in result.iterrows():
            # logger.debug(f"Row: {row}")
            message = MessageInfo(
                message_id=row["message_id"],
                user=row["user"],
                platform=row["platform"],
                llm_name=row["llm_name"],
                query=row["query"],
                answer=row["answer"],
                filters=json.loads(row["filters"]) if row["filters"] else None,
                sources=json.loads(row["sources"]) if row["sources"] else None,
                created_at=row["created_at"],
                updated_at=row["created_at"],
                llm_context=json.loads(row["llm_context"]) if row["llm_context"] else None,
                generated_media=json.loads(row["generated_media"]) if row["generated_media"] else None,
                feedback=json.loads(row["feedback"]) if row["feedback"] else None,
                conversation_id=row["conversation_id"],
                state=row["state"],
                # history=json.loads(row["history"]) if row["history"] else None,
            )
            messages.append(message)
        return messages

    @log_query_time
    def get_message(self, user: str, message_id: str):
        conditions = [
            WhereCondition(
                column=self.col("message_id"),
                value=message_id,
                operator=Operator.EQ,
            ),
            WhereCondition(column=self.col("user"), value=user, operator=Operator.EQ),
        ]
        return self.select_columns_from_dataset(column_names=self.columns, eq_cond=conditions)

    @log_query_time
    def delete_conversation_messages(self, user: str, conversation_id: str):
        conditions = [
            WhereCondition(
                column=self.col("conversation_id"),
                value=conversation_id,
                operator=Operator.EQ,
            ),
            WhereCondition(column=self.col("user"), value=user, operator=Operator.EQ),
        ]
        self.delete_files(conditions)
        if self.permanent_delete:
            self.delete_messages_permanently(conditions)
        else:
            self.update_messages_as_deleted(conditions)


messages_sql_manager = MessageSQL()
