import json
import sys

from answers.backend.db.conversations import conversations_sql_manager
from answers.backend.db.messages import messages_sql_manager
from answers.backend.db.sql.dataset_schemas import CONVERSATION_DATASET_CONF_ID, MESSAGE_DATASET_CONF_ID
from answers.backend.utils.db.conversation_utils import handle_images
from answers.solutions.service import LLM_Question_Answering
from common.backend.schemas.ask import AskRequestSchema, AskResponseDataSchema
from common.backend.services.ask import AskService
from common.backend.utils.dataiku_api import dataiku_api
from common.llm_assist.llm_api_handler import llm_setup
from common.llm_assist.logging import logger
from flask import Blueprint, Response, jsonify, request, stream_with_context

ask_bp = Blueprint("ask", __name__)


def str_to_bool(value):
    """Convert query string value to boolean."""
    return value.lower() in ["true", "1", "yes", "on"]


@ask_bp.route("/ask", methods=["POST"])
def process_query():
    config = dataiku_api.webapp_config
    enable_api = config.get("enable_answers_api", False)
    messages_dataset = config.get(MESSAGE_DATASET_CONF_ID)
    if not enable_api or not messages_dataset:
        logger.debug("Ask service not enabled.")
        return jsonify(error="Ask service not available."), 500
    enable_conversation = config.get(CONVERSATION_DATASET_CONF_ID, None) is not None
    accept_header = request.headers.get("Accept", "text/event-stream").lower()

    if accept_header not in ["text/event-stream", "application/json"]:
        return jsonify(error="Not Acceptable: Supported types are text/event-stream, application/json."), 406

    streaming = accept_header == "text/event-stream"

    # TODO validate auth ?
    # Load request data and validate using schema
    request_data = request.get_json()
    schema = AskRequestSchema()
    errors = schema.validate(request_data)
    if errors:
        logger.debug(f"Validation errors: {errors}")
        return jsonify(errors), 400

    llm_qa = LLM_Question_Answering(llm_setup.get_llm())

    if streaming:

        def stream_answer():
            try:
                for chunk in AskService.process_query_streaming(
                    request_data=request_data,
                    llm_qa=llm_qa,
                    messages_sql_manager=messages_sql_manager,
                    conversation_sql_manager=conversations_sql_manager if enable_conversation else None,
                    handle_images=handle_images,
                ):
                    yield f"event: {chunk.event}"
                    yield "\n"
                    yield f"data: {json.dumps(chunk.data)}"
                    yield "\n\n"
                    yield ": flush\n\n"
                    sys.stdout.flush()

            except Exception as e:
                logger.exception(f"Error processing query: {e}")
                yield f"event: completion-end"
                yield "\n"
                yield f"data: {json.dumps({'status': 'ko', 'message': 'Error processing your request'})}"
                yield "\n\n"
                return

        return Response(
            stream_with_context(stream_answer()),
            mimetype="text/event-stream; charset=utf-8",
            headers={
                "Cache-Control": "no-cache",
                "X-Accel-Buffering": "no",
                "Connection": "keep-alive",
            },
        )
    else:
        try:
            processed_data = AskService.process_query(
                data=request_data,
                llm_qa=llm_qa,
                messages_sql_manager=messages_sql_manager,
                conversation_sql_manager=conversations_sql_manager if enable_conversation else None,
                handle_images=handle_images,
            )

            # Build and validate the response using the schema
            response_schema = AskResponseDataSchema()
            validated_response = response_schema.dump(processed_data)
            # Return the response
            return jsonify({"status": "ok", "data": validated_response})
        except Exception as e:
            logger.exception(f"Error processing query: {e}")
            return jsonify({"status": "ko", "message": "Error processing your request"}), 500
