from functools import lru_cache
from typing import Any, Dict, List, Optional, TypedDict, Union

import dataiku
import pandas as pd
from answers.solutions.vector_search.vector_query import VectorSearch, VectorStoreTypes
from common.backend.utils.dataiku_api import dataiku_api
from common.backend.utils.dataiku_utils import find_recipe
from common.backend.utils.metas_utils import convert_to_list, is_string_list_representation
from common.backend.utils.sql_timing import log_query_time
from common.llm_assist.logging import logger


class KnowledgeBankFilterConfig(TypedDict):
    input_dataset: str
    filter_columns: List[str]
    filter_options: Dict[str, List[Any]]
    vector_db_type: Optional[str]


def flatten(list_of_lists):
    return [item for sublist in list_of_lists for item in sublist]


# Disable lists flattening for now, just show what has been configured
manage_lists = False


@log_query_time
def compute_filter_options(dataset: dataiku.Dataset, columns: List[str]) -> Dict[str, List[Any]]:
    df: pd.DataFrame = dataset.get_dataframe(columns=columns, parse_dates=False, infer_with_pandas=False)
    result = {}
    for column in df.columns:
        if manage_lists:
            if df[column].apply(lambda x: isinstance(x, str) and is_string_list_representation(x)).any():
                df[column] = df[column].apply(convert_to_list)

            if df[column].apply(lambda x: isinstance(x, list)).any():
                flattened_list = [item for sublist in df[column].dropna().tolist() for item in sublist]
                unique_values = pd.unique(pd.Series(flattened_list)).tolist()
            else:
                unique_values = pd.unique(df[column].dropna()).tolist()
        else:
            unique_values = pd.unique(df[column].dropna()).tolist()

        result[column] = unique_values
    return result



@log_query_time
@lru_cache(maxsize=None)
def get_knowledge_bank_filtering_settings(knowledge_bank_id: str, with_options: bool = True):
    project = dataiku_api.default_project
    config = dataiku_api.webapp_config
    vector_db_type: Optional[str] = None
    if not with_options and knowledge_bank_id:
        vector_db_type = (
            project.get_knowledge_bank(knowledge_bank_id).as_core_knowledge_bank()._get()["vectorStoreType"]
        )

    flow = project.get_flow()
    graph = flow.get_graph()
    recipe_json = None
    try:
        recipe_json = find_recipe(graph.data, knowledge_bank_id)
    except KeyError:
        raise KeyError(knowledge_bank_id)

    dataset_name: str = recipe_json.get("predecessors")[0]
    dataset = dataiku.Dataset(project_key=project.project_key, name=dataset_name)
    recipe_name = recipe_json.get("ref")
    recipe = project.get_recipe(recipe_name)
    recipe_settings = recipe.get_settings()
    recipe_json_payload = recipe_settings.get_json_payload()
    kb_columns = []
    if "metadataColumns" in recipe_json_payload:
        metadataColumns = recipe_json_payload["metadataColumns"]
        if metadataColumns:
            for metadata in metadataColumns:
                if with_options:
                    if metadata["column"] in list(config.get("knowledge_sources_filters", [])):
                        kb_columns.append(metadata["column"])
                else:
                    kb_columns.append(metadata["column"])

    filter_options = {}
    if len(kb_columns) > 0 and with_options:
        filter_options = compute_filter_options(dataset, columns=kb_columns)

    return KnowledgeBankFilterConfig(
        input_dataset=dataset_name,
        filter_columns=kb_columns,
        filter_options=filter_options,
        vector_db_type=vector_db_type,
    )


def get_current_filter_config():
    # TODO: Adapt when several knowledge banks will be connected
    knowledge_bank_id = dataiku_api.webapp_config.get("knowledge_bank_id", None)
    if knowledge_bank_id:
        try:
            result = get_knowledge_bank_filtering_settings(knowledge_bank_id)
            return result
        except KeyError:
            return None
    return None


def filters_to_dsl(filters: Dict[str, List[Any]]) -> Dict[str, Any]:
    if len(filters) < 1:
        raise ValueError("No filters provided")
    elif len(filters) == 1:
        col, values = list(filters.items())[0]
        return {"filter" : {col : {"in": values}}}
    return {"filter" : {"and": [{col : {"in": values}} for col, values in filters.items()]}}


def process_filters_for_db(filters: Union[Dict[str, List[Any]], None], vector_db_type):
    if filters is None:
        return None
    logger.debug(f"processing filters for {vector_db_type}")
    if vector_db_type not in dir(VectorStoreTypes):
        raise ValueError(f"Vector DB type {vector_db_type} is unknown")
    vec_search = VectorSearch(vector_store_type=vector_db_type).get_vector_store()
    processed_filters =  vec_search.invoke(filters_to_dsl(filters))
    return processed_filters
