import inspect
import json
from typing import Any, Dict, List, Optional, Tuple, cast

from answers.solutions.knowledge_bank import get_core_knowledge_bank, get_knowledge_bank_infos
from common.backend.utils.dataiku_api import dataiku_api
from common.backend.utils.llm_utils import get_llm_capabilities
from common.solutions.prompts.conversations import ChatHistoryHandler
from dataiku.core.knowledge_bank import MultipartContext
from langchain.callbacks.manager import CallbackManagerForChainRun
from langchain.chains import ConversationalRetrievalChain
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.schema.document import Document
from langchain_core.prompt_values import PromptValue


class DKUConversationRetrievalChain(ConversationalRetrievalChain):
    """DKUConversationRetrievalChain is a subclass of ConversationalRetrievalChain
    designed to tailor the conversational retrieval process with a specific
    implementation for generating the final prompt based on chat history and
    question modification, to be able to use it for the streaming
    It works only when the chain_type is "stuff"
    """

    def _extract_docs(
        self, inputs: Dict[str, Any], question: str, run_manager: CallbackManagerForChainRun
    ) -> List[Document]:
        """Extract the documents from the inputs"""
        docs: List[Document]
        accepts_run_manager = "run_manager" in inspect.signature(self._get_docs).parameters
        if accepts_run_manager:
            docs = self._get_docs(question, inputs, run_manager=run_manager)
        else:
            docs = self._get_docs(question, inputs)  # type: ignore[call-arg]
        return docs

    def _prepare_llm_inputs(
        self, inputs: Dict[str, Any], docs, chat_history_str: Optional[str], is_multimodal: bool = False
    ) -> Tuple[MultipartContext, Dict[str, Any]]:
        """Build new inputs for the llm chain"""
        new_inputs = inputs.copy()
        # new_inputs["question"] = question
        new_inputs["chat_history"] = chat_history_str
        # Get the inputs is only possible for chain_type "stuff"

        kb_id = dataiku_api.webapp_config.get("knowledge_bank_id")
        _, knowledge_bank_schema = get_knowledge_bank_infos(kb_id)

        multipart_context = None
        final_inputs: Dict[str, Any] = {}
        if is_multimodal and get_llm_capabilities().get("multi_modal", False):
            kb = get_core_knowledge_bank(kb_id)
            multipart_context = MultipartContext()
            for doc in docs:
                doc.metadata["parts"] = []
                mpc = kb.get_multipart_context([doc])
                for part in mpc.parts:
                    if part.type == "IMAGE_INLINE":
                        doc.metadata["parts"].append(part)
                multipart_context.parts.extend(mpc.parts)
            final_inputs["context"] = ""
        else:
            documents_context = self.from_retriever_documents_to_structured_context(docs, knowledge_bank_schema)
            final_inputs["context"] = documents_context

        final_inputs["chat_history"] = chat_history_str
        return multipart_context, final_inputs

    def prepare_final_answer_prompt(
        self,
        condensed_query: str,
        kb_query: str,
        inputs: Dict[str, Any],
        run_manager: Optional[CallbackManagerForChainRun] = None,
        is_multimodal: bool = False,
    ) -> Tuple[MultipartContext, PromptValue, Dict[str, Any]]:
        """Prepares a prompt for answer generation by using chat history to refine
        the question and incorporating relevant documents for context.

        Parameters
        ----------
        - inputs (dictionary): Contains the initial question and chat history.
        - run_manager (CallbackManagerForChainRun, optional): Manages callbacks, defaults to a no-op manager.

        Returns
        -------
        - (List[PromptValue], dictionary): The final prompt(s) and additional output details like 'source_documents'.

        Note:
        - Optimized for use in conversational retrieval chains where chat history impacts question formation and answer generation.
        - Accommodates both modified and unchanged questions influenced by chat history.

        """
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        chat_history_handler = ChatHistoryHandler()
        chat_history_str = chat_history_handler.format_chat_history(
            chat_history=inputs["chat_history"], dialogs_format="tuples"
        )
        # Extract docs
        docs = self._extract_docs(inputs, question=kb_query, run_manager=_run_manager)
        multipart_context, final_inputs = self._prepare_llm_inputs(
            inputs,
            docs=docs,
            chat_history_str=chat_history_str,
            is_multimodal=is_multimodal,
        )
        stuff_chain = cast(StuffDocumentsChain, self.combine_docs_chain)
        final_prompt = stuff_chain.llm_chain.prep_prompts(input_list=[final_inputs], run_manager=_run_manager)  # type: ignore
        # Prepare the output
        question_info: Dict[str, Any] = {}

        if self.return_source_documents:
            question_info["source_documents"] = docs

        if self.return_generated_question:
            question_info["generated_question"] = condensed_query
        return multipart_context, final_prompt[0][0], question_info

    def from_retriever_documents_to_structured_context(
        self, list_of_documents: List[Document], knowledge_bank_schema
    ) -> str:
        structured_documents = []

        metadata_fields = dataiku_api.webapp_config.get("knowledge_sources_context_metadata")

        for document_index, document in enumerate(list_of_documents):
            if knowledge_bank_schema and metadata_fields:
                filtered_metadata = [
                    {
                        "description": schema.get("comment", ""),
                        "name": schema["name"],
                        "value": document.metadata[schema["name"]],
                    }
                    for schema in knowledge_bank_schema
                    if schema["name"] in document.metadata and schema["name"] in metadata_fields
                ]

                structured_document = {
                    "Source number": document_index + 1,
                    "Content": document.page_content,
                    "Metadata": filtered_metadata,
                }
            else:
                structured_document = {
                    "Source number": document_index + 1,
                    "Content": document.page_content,
                }
            structured_documents.append(structured_document)

        documents_context = json.dumps(structured_documents, ensure_ascii=False, indent=4)
        return documents_context
