import json
from typing import Any, Dict, List, Union

from common.backend.utils.langchain_utils import escape_dictionary_for_langchain


def load_json_query_structure_explanations(json_sql_key_explanations: Dict[str, str], forbidden_keys: List[str], json_sql_key_types: Dict[str, str])->str:
    """Loads the relevant explanations do describe the structure of the JSONs allowing to query databases.
    Explanations are filtered leveraging the constant 'JSON_SQL_FORBIDDEN_KEYS'

    :returns: json_query_structure_explanations: str: The JSONs structure explanation.
    """
    json_query_structure_explanations = ""
    keys_to_explain = [
        key for key in json_sql_key_explanations.keys()
        if key not in forbidden_keys
    ]
    for key_index, key in enumerate(keys_to_explain):
        key_explanation = json_sql_key_explanations[key]
        json_query_structure_explanations += f"- {key_index+1}. '{key}' {json_sql_key_types[key]}: {key_explanation}"
    json_query_structure_explanations += f"- {key_index+2}. 'justification' [STRING]: "\
                                         "Explanation of why chose to use or not use this query."
    return json_query_structure_explanations


def load_empty_json_string(json_sql_key_explanations: Dict[str, str], forbidden_keys:  List[str])-> Dict[str, Union[str, List]]:
    """Loads the string explaining the 'empty JSON' that should be generated by the LLM when 
    it detects that no query should be passed to the database.

    :returns: empty_json_string: str: The 'empty JSON' example of the LLM.
    """
    empty_json: Dict[str, Union[str, list, None]] = {}
    keys_to_explain = [
        key for key in json_sql_key_explanations.keys()
        if key not in forbidden_keys
    ]
    for key in keys_to_explain:
        empty_json[key] = None
    empty_json["justification"] = "<YOUR JUSTIFICATION>"
    empty_json_keys = list(empty_json.keys())
    empty_json_string = escape_dictionary_for_langchain(empty_json)
    return {"empty_json_string": empty_json_string, 
            "empty_json_keys": empty_json_keys}


def load_sql_examples_prompt(raw_sql_conversation: list,
                             forbidden_keys: list)->str:
    """Loads the prompt associated with the few-shot examples explaining how the LLM should
    generate JSONs based on the available datasource and the questions asked by a 'Human'.
    
    :param datasource_description: str: The description of the available datasource, as get from the function 'get_dataset_descriptions'.
    :param raw_sql_conversation: list: A list containing conversation turns between a 'Human' and an 'Assistant'. 
        It has the format [
            {'Human': HUMAN_QUESTION_1, 'Assistant': ASSISTANT_GENERATED_JSON_1},
            {'Human': HUMAN_QUESTION_2, 'Assistant': ASSISTANT_GENERATED_JSON_2},
                ...
            {'Human': HUMAN_QUESTION_N, 'Assistant': ASSISTANT_GENERATED_JSON_N}
        ]
    :param forbidden_keys: list: The list of the keys that are forbidden in the JSONs generated by the Assistant.

    :returns: sql_examples_prompt: str: The prompt containing the examples of how to behave.
    """
    sql_examples_prompt = ""
    for conversation_index, conversation in enumerate(raw_sql_conversation):
        human_question = conversation["Human"]
        assistant_answer = conversation["Assistant"]
        assistant_answer = {key: value for key, value in assistant_answer.items()
                            if key not in forbidden_keys}
        
        human_question = escape_dictionary_for_langchain(json.dumps(human_question))
        assistant_answer = escape_dictionary_for_langchain(json.dumps(assistant_answer))
        
        
        sql_examples_prompt += f"### Example {conversation_index+1}\n"\
        f"[HumanMessage(content='{human_question}'), \n"\
        f"AIMessage(content='{escape_dictionary_for_langchain(assistant_answer)}')]\n\n"
    return sql_examples_prompt

def structure_examples(example_list: List[Dict[str, Any]], json_sql_forbidden_keys: List[str]) -> str:
    formatted_examples = []
    for example in example_list:
        data_description = example['data_description']
        full_example_str = load_sql_examples_prompt(example['example_conversations'], json_sql_forbidden_keys)
        
        formatted_examples.append(f"""### SQL Table Description and Schema
        {data_description}
        ### Expected example responses
        {full_example_str}
        --------------------------------            
        """)
    return '\n'.join(formatted_examples)
    

