import json
import re
from typing import Any, Dict, Optional

from common.llm_assist.logging import logger


def extract_json(response_text: str, json_pattern: Optional[str]) -> Dict[str, Any]:
    # Find all characters that could be the start or end of a JSON object
    json_objects = re.findall(json_pattern if json_pattern else r"{[\s\S]*?}", response_text)
    # Find the longest string that could be a JSON object,
    # since it's most likely to be the correct one
    longest_json = max(json_objects, key=len, default="{}")
    # Convert the string back to a dictionary (JSON object)
    try:
        json_data: Dict[str, Any] = json.loads(longest_json)
    except json.JSONDecodeError as e:
        logger.exception("Error parsing json")
        # In case JSON decoding fails, just return an empty dictionary
        # TODO - This should be handled better maybe no more need to extract longest json
        json_data = json.loads(response_text)

    return json_data


def remove_keys_from_json(json_obj, keys_to_remove):
    """
    Recursively remove multiple keys from a JSON object (dict or list).
    """
    if isinstance(json_obj, dict):
        return {
            k: remove_keys_from_json(v, keys_to_remove)
            for k, v in json_obj.items()
            if k not in keys_to_remove
        }
    elif isinstance(json_obj, list):
        return [remove_keys_from_json(item, keys_to_remove) for item in json_obj]
    else:
        return json_obj

def strip_keys_from_json(json_str, keys_to_remove):
    """
    Takes a JSON string, removes all occurrences of the given keys, and returns the modified JSON string.
    """
    try:
        json_obj = json.loads(json_str)
        cleaned_json = remove_keys_from_json(json_obj, keys_to_remove)
        return json.dumps(cleaned_json, indent=2)
    except json.JSONDecodeError:
        return "Invalid JSON"
    
def remove_keys_from_dict(data, keys_to_remove):
    """
    Recursively removes specified keys from a dictionary.
    """
    if isinstance(data, dict):
        return {
            k: remove_keys_from_dict(v, keys_to_remove)
            for k, v in data.items()
            if k not in keys_to_remove
        }
    elif isinstance(data, list):
        return [remove_keys_from_dict(item, keys_to_remove) for item in data]
    else:
        return data