
from abc import ABC, abstractmethod
from typing import Any, Dict

from common.backend.constants import DEFAULT_DECISIONS_GENERATION_ERROR
from common.backend.utils.json_utils import extract_json
from common.backend.utils.llm_utils import (
    get_alternative_llm,
    get_llm_completion,
    handle_response_trace,
    parse_error_messages,
)
from common.backend.utils.sql_timing import log_query_time
from common.llm_assist.logging import logger
from dataiku.langchain.dku_llm import DKULLM
from dataikuapi.dss.llm import DSSLLMCompletionQuery, DSSLLMCompletionResponse


class GenericDecisionJSONChain(ABC):
    # Abstract class for Decision Chain that returns JSON objects
    # Implement the prompt property in the child class
    @property
    @abstractmethod
    # The prompt that will be used to generate the decision
    # This should instruct the model on what decision to make
    # and what information to include in the decision
    # This should instruct the model to output JSON!!!!
    def prompt(self):
        raise NotImplementedError("Subclasses must implement prompt property")

    @property
    @abstractmethod
    def llm(self):
        raise NotImplementedError("Subclasses must implement llm property")

    @property
    @abstractmethod
    def chain_purpose(self) -> str:
        raise NotImplementedError("Subclasses must implement chain_purpose property")

    @abstractmethod
    def verified_json_output(self, json_output: Dict[str, Any]) -> Dict[str, Any]:
        pass

    def get_decision_json_pattern(self) -> str:
        return r"{[\s\S]*?}"
    
    @log_query_time
    def get_decision_as_json(self, user_query: str) -> Dict[str, Any]:
        error_message = DEFAULT_DECISIONS_GENERATION_ERROR
        if self.prompt is None:
            logger.error("Prompt is not defined")
            raise Exception("Decision prompt is not defined")
        logger.debug(f"{self.chain_purpose} Final Prompt: {self.prompt}")
        decision_llm: DKULLM = get_alternative_llm("json_decision_llm_id")
        completion: DSSLLMCompletionQuery = get_llm_completion(decision_llm)
        try:
            completion.with_message(self.prompt, role="system")
            completion.with_message(user_query, role="user")
            llm_response: DSSLLMCompletionResponse = completion.execute()
            handle_response_trace(llm_response)
            response: str = llm_response.text
            logger.debug(f"{self.chain_purpose} llm response: '{response}'")
            if not response and llm_response.errorMessage:
                error_message += str(llm_response.errorMessage)
                raise Exception(error_message)
            if not response:
                error_message += "No response from LLM"
                raise Exception(error_message)
            if isinstance(response, dict):
                json_response = response
            else:
                json_response = extract_json(response, json_pattern=self.get_decision_json_pattern())

            logger.debug(f"Extracted response  {json_response}")
            verified_resp = self.verified_json_output(json_response)
            return verified_resp
        except Exception as e:
            error_message += parse_error_messages(str(e)) or str(e)
            logger.exception(error_message)
            raise Exception(error_message)